"""Public async Python client built on top of the protobuf transport."""

from __future__ import annotations

import asyncio
import dataclasses
import os
import platform
from collections.abc import AsyncIterator, Callable, Mapping, Sequence
from pathlib import Path
from threading import Event as ThreadEvent
from threading import Thread
from typing import cast

from ._generated import service_pb2
from ._grpc_client import SandboxGrpcClient
from ._conversions import (
    normalize_from_sequence,
    parse_event_sequence,
    to_exec_handle,
    to_exec_snapshot,
    to_ping_info,
    to_proto_create_exec_request,
    to_proto_create_sandbox_request,
    to_sandbox_event,
    to_sandbox_handle,
)
from .errors import (
    ExecNotRunningError,
    SandboxClientError,
    SandboxConflictError,
    SandboxInvalidStateError,
)
from .models import SandboxState
from ._request_types import CreateExecRequest, CreateSandboxRequest, CreateSandboxSpec
from .types import (
    CopySpec,
    DeleteSandboxesResult,
    ExecHandle,
    MountSpec,
    PingInfo,
    SandboxEvent,
    SandboxHandle,
    ServiceSpec,
)


def _resolve_default_socket_path(
    *,
    system: str | None = None,
    lookup_env: Callable[[str], str | None] | None = None,
    home_dir: str | None = None,
) -> str:
    lookup = os.environ.get if lookup_env is None else lookup_env
    resolved_system = platform.system() if system is None else system
    if resolved_system == "Darwin":
        resolved_home = home_dir or os.path.expanduser("~")
        if resolved_home:
            return os.path.join(
                resolved_home,
                "Library",
                "Application Support",
                "agbox",
                "run",
                "agboxd.sock",
            )
    else:
        runtime_dir = lookup("XDG_RUNTIME_DIR")
        if runtime_dir:
            return os.path.join(runtime_dir, "agbox", "agboxd.sock")
        raise RuntimeError("XDG_RUNTIME_DIR is required to resolve the AgentsSandbox socket path on Linux")


def _validate_optional_id(field_name: str, value: str | None) -> str | None:
    if value == "":
        raise ValueError(f"{field_name} must not be empty")
    return value


class AgentsSandboxClient:
    """Async high-level client that exposes the northbound SDK surface."""

    def __init__(
        self,
        *,
        timeout_seconds: float = 5.0,
        stream_timeout_seconds: float | None = None,
        operation_timeout_seconds: float = 60.0,
    ) -> None:
        self.socket_path = _resolve_default_socket_path()
        self._timeout_seconds = timeout_seconds
        self._stream_timeout_seconds = stream_timeout_seconds
        self._operation_timeout_seconds = operation_timeout_seconds
        self._rpc_client = SandboxGrpcClient(self.socket_path, timeout_seconds=timeout_seconds)

    def close(self) -> None:
        self._rpc_client.close()

    async def __aenter__(self) -> AgentsSandboxClient:
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        del exc_type, exc, tb
        self.close()

    async def ping(self) -> PingInfo:
        response = await asyncio.to_thread(self._rpc_client.ping)
        return to_ping_info(response)

    async def create_sandbox(
        self,
        *,
        config: str | Path | None = None,
        image: str | None = None,
        sandbox_id: str | None = None,
        mounts: tuple[MountSpec, ...] = (),
        copies: tuple[CopySpec, ...] = (),
        builtin_resources: tuple[str, ...] = (),
        required_services: tuple[ServiceSpec, ...] = (),
        optional_services: tuple[ServiceSpec, ...] = (),
        labels: Mapping[str, str] | None = None,
        wait: bool = True,
    ) -> SandboxHandle:
        if config is None and image is None:
            raise ValueError("at least one of 'config' or 'image' must be provided")

        config_yaml = b""
        if config is not None:
            config_yaml = Path(config).read_bytes()

        request = CreateSandboxRequest(
            sandbox_id=_validate_optional_id("sandbox_id", sandbox_id),
            create_spec=CreateSandboxSpec(
                image=image,
                mounts=mounts,
                copies=copies,
                builtin_resources=builtin_resources,
                required_services=required_services,
                optional_services=optional_services,
                labels={} if labels is None else dict(labels),
            ),
            config_yaml=config_yaml,
        )
        try:
            response = await asyncio.to_thread(
                self._rpc_client.create_sandbox,
                to_proto_create_sandbox_request(request),
            )
        except SandboxConflictError as exc:
            if request.sandbox_id is None:
                raise
            raise SandboxConflictError(f"Sandbox already exists for id {request.sandbox_id}.") from exc
        current = await self.get_sandbox(response.sandbox_id)
        if not wait:
            return current
        return await self._wait_for_sandbox_state(
            sandbox_id=response.sandbox_id,
            baseline=current,
            target_state=SandboxState.READY,
            operation_name="create_sandbox",
        )

    async def get_sandbox(self, sandbox_id: str) -> SandboxHandle:
        response = await asyncio.to_thread(self._rpc_client.get_sandbox, sandbox_id)
        return to_sandbox_handle(response.sandbox)

    async def list_sandboxes(
        self,
        *,
        include_deleted: bool = False,
        label_selector: Mapping[str, str] | None = None,
    ) -> list[SandboxHandle]:
        response = await asyncio.to_thread(
            self._rpc_client.list_sandboxes,
            service_pb2.ListSandboxesRequest(
                include_deleted=include_deleted,
                label_selector={} if label_selector is None else dict(label_selector),
            ),
        )
        return [to_sandbox_handle(item) for item in response.sandboxes]

    async def resume_sandbox(
        self,
        sandbox_id: str,
        *,
        wait: bool = True,
    ) -> SandboxHandle:
        await asyncio.to_thread(self._rpc_client.resume_sandbox, sandbox_id)
        current = await self.get_sandbox(sandbox_id)
        if not wait:
            return current
        return await self._wait_for_sandbox_state(
            sandbox_id=sandbox_id,
            baseline=current,
            target_state=SandboxState.READY,
            operation_name="resume_sandbox",
        )

    async def stop_sandbox(
        self,
        sandbox_id: str,
        *,
        wait: bool = True,
    ) -> SandboxHandle:
        await asyncio.to_thread(self._rpc_client.stop_sandbox, sandbox_id)
        current = await self.get_sandbox(sandbox_id)
        if not wait:
            return current
        return await self._wait_for_sandbox_state(
            sandbox_id=sandbox_id,
            baseline=current,
            target_state=SandboxState.STOPPED,
            operation_name="stop_sandbox",
        )

    async def delete_sandbox(
        self,
        sandbox_id: str,
        *,
        wait: bool = True,
    ) -> SandboxHandle:
        await asyncio.to_thread(self._rpc_client.delete_sandbox, sandbox_id)
        current = await self.get_sandbox(sandbox_id)
        if not wait:
            return current
        return await self._wait_for_sandbox_state(
            sandbox_id=sandbox_id,
            baseline=current,
            target_state=SandboxState.DELETED,
            operation_name="delete_sandbox",
        )

    async def delete_sandboxes(
        self,
        label_selector: Mapping[str, str],
        *,
        wait: bool = True,
    ) -> DeleteSandboxesResult:
        response = await asyncio.to_thread(
            self._rpc_client.delete_sandboxes,
            service_pb2.DeleteSandboxesRequest(label_selector=dict(label_selector)),
        )
        result = DeleteSandboxesResult(
            deleted_sandbox_ids=tuple(response.deleted_sandbox_ids),
            deleted_count=response.deleted_count,
        )
        if not wait or not result.deleted_sandbox_ids:
            return result
        baselines = await asyncio.gather(*(self.get_sandbox(sandbox_id) for sandbox_id in result.deleted_sandbox_ids))
        await asyncio.gather(
            *(
                self._wait_for_sandbox_state(
                    sandbox_id=sandbox_id,
                    baseline=baseline,
                    target_state=SandboxState.DELETED,
                    operation_name="delete_sandboxes",
                )
                for sandbox_id, baseline in zip(result.deleted_sandbox_ids, baselines, strict=True)
            )
        )
        return result

    async def create_exec(
        self,
        sandbox_id: str,
        command: Sequence[str],
        *,
        exec_id: str | None = None,
        cwd: str = "/workspace",
        env_overrides: Mapping[str, str] | None = None,
        wait: bool = False,
    ) -> ExecHandle:
        request = CreateExecRequest(
            sandbox_id=sandbox_id,
            command=tuple(command),
            exec_id=_validate_optional_id("exec_id", exec_id),
            cwd=cwd,
            env_overrides={} if env_overrides is None else dict(env_overrides),
        )
        response = await asyncio.to_thread(
            self._rpc_client.create_exec,
            to_proto_create_exec_request(request),
        )
        stdout_log_path = response.stdout_log_path or None
        stderr_log_path = response.stderr_log_path or None
        current, last_event_sequence = await self._get_exec_snapshot(response.exec_id)
        current = dataclasses.replace(
            current,
            stdout_log_path=stdout_log_path,
            stderr_log_path=stderr_log_path,
        )
        if not wait:
            return current
        handle = await self._wait_for_exec_terminal(
            exec_id=response.exec_id,
            sandbox_id=sandbox_id,
            baseline=current,
            baseline_sequence=last_event_sequence,
            operation_name="create_exec",
        )
        return dataclasses.replace(
            handle,
            stdout_log_path=stdout_log_path,
            stderr_log_path=stderr_log_path,
        )

    async def run(
        self,
        sandbox_id: str,
        command: Sequence[str],
        *,
        cwd: str = "/workspace",
        env_overrides: Mapping[str, str] | None = None,
    ) -> ExecHandle:
        return await self.create_exec(
            sandbox_id,
            command,
            cwd=cwd,
            env_overrides=env_overrides,
            wait=True,
        )

    async def cancel_exec(
        self,
        exec_id: str,
        *,
        wait: bool = True,
    ) -> ExecHandle:
        try:
            await asyncio.to_thread(self._rpc_client.cancel_exec, exec_id)
        except SandboxInvalidStateError as exc:
            raise ExecNotRunningError(exec_id) from exc
        current, last_event_sequence = await self._get_exec_snapshot(exec_id)
        if not wait:
            return current
        return await self._wait_for_exec_terminal(
            exec_id=exec_id,
            sandbox_id=current.sandbox_id,
            baseline=current,
            baseline_sequence=last_event_sequence,
            operation_name="cancel_exec",
        )

    async def get_exec(self, exec_id: str) -> ExecHandle:
        handle, _ = await self._get_exec_snapshot(exec_id)
        return handle

    async def list_active_execs(
        self,
        sandbox_id: str | None = None,
    ) -> list[ExecHandle]:
        response = await asyncio.to_thread(
            self._rpc_client.list_active_execs,
            "" if sandbox_id is None else sandbox_id,
        )
        return [to_exec_handle(item) for item in response.execs]

    async def _get_exec_snapshot(self, exec_id: str) -> tuple[ExecHandle, int]:
        response = await asyncio.to_thread(self._rpc_client.get_exec, exec_id)
        return to_exec_snapshot(response)

    async def subscribe_sandbox_events(
        self,
        sandbox_id: str,
        *,
        from_sequence: int = 0,
        include_current_snapshot: bool = False,
    ) -> AsyncIterator[SandboxEvent]:
        normalized_sequence = normalize_from_sequence(from_sequence)
        loop = asyncio.get_running_loop()
        queue: asyncio.Queue[tuple[str, SandboxEvent | Exception | None]] = asyncio.Queue()
        stop_event = ThreadEvent()
        stream_client = SandboxGrpcClient(
            self.socket_path,
            timeout_seconds=(
                self._timeout_seconds if self._stream_timeout_seconds is None else self._stream_timeout_seconds
            ),
        )

        def publish(kind: str, payload: SandboxEvent | Exception | None) -> None:
            try:
                loop.call_soon_threadsafe(queue.put_nowait, (kind, payload))
            except RuntimeError:
                return

        def produce() -> None:
            try:
                for item in stream_client.subscribe_sandbox_events(
                    sandbox_id,
                    from_sequence=normalized_sequence,
                    include_current_snapshot=include_current_snapshot,
                ):
                    if stop_event.is_set():
                        break
                    publish("event", to_sandbox_event(item))
            except Exception as exc:  # noqa: BLE001
                if not stop_event.is_set():
                    publish("error", exc)
            finally:
                stream_client.close()
                publish("end", None)

        producer = Thread(target=produce, name=f"agents-sandbox-events-{sandbox_id}", daemon=True)
        producer.start()
        try:
            while True:
                kind, payload = await queue.get()
                if kind == "event":
                    yield cast(SandboxEvent, payload)
                    continue
                if kind == "error":
                    raise cast(Exception, payload)
                break
        finally:
            stop_event.set()
            stream_client.close()
            await asyncio.to_thread(producer.join, 1.0)

    async def _wait_for_sandbox_state(
        self,
        *,
        sandbox_id: str,
        baseline: SandboxHandle,
        target_state: SandboxState,
        operation_name: str,
    ) -> SandboxHandle:
        if baseline.state == target_state:
            return baseline
        self._raise_for_failed_sandbox(baseline, operation_name)
        baseline_sequence = parse_event_sequence(baseline.last_event_sequence)
        stream_sequence = baseline.last_event_sequence
        try:
            async with asyncio.timeout(self._operation_timeout_seconds):
                async for event in self.subscribe_sandbox_events(sandbox_id, from_sequence=stream_sequence):
                    if event.sequence <= baseline_sequence:
                        continue
                    current = await self.get_sandbox(sandbox_id)
                    if current.state == target_state:
                        return current
                    self._raise_for_failed_sandbox(current, operation_name)
        except TimeoutError as exc:
            raise TimeoutError(
                f"{operation_name} timed out while waiting for sandbox {sandbox_id} to reach {target_state.name}."
            ) from exc
        raise SandboxClientError(
            f"{operation_name} ended before sandbox {sandbox_id} reached {target_state.name}."
        )

    async def _wait_for_exec_terminal(
        self,
        *,
        exec_id: str,
        sandbox_id: str,
        baseline: ExecHandle,
        baseline_sequence: int,
        operation_name: str,
    ) -> ExecHandle:
        if baseline.state.is_terminal:
            return baseline
        baseline_sequence = parse_event_sequence(baseline_sequence)
        stream = self.subscribe_sandbox_events(sandbox_id, from_sequence=baseline_sequence)
        try:
            async with asyncio.timeout(self._operation_timeout_seconds):
                async for event in stream:
                    if event.sequence <= baseline_sequence:
                        continue
                    if event.exec_id != exec_id:
                        continue
                    current, _ = await self._get_exec_snapshot(exec_id)
                    if current.state.is_terminal:
                        return current
        except TimeoutError as exc:
            raise TimeoutError(
                f"{operation_name} timed out while waiting for exec {exec_id} to become terminal."
            ) from exc
        finally:
            await stream.aclose()
        raise SandboxClientError(
            f"{operation_name} event stream ended before exec {exec_id} reached a terminal state."
        )

    def _raise_for_failed_sandbox(self, handle: SandboxHandle, operation_name: str) -> None:
        if handle.state == SandboxState.FAILED:
            raise SandboxClientError(
                f"{operation_name} observed sandbox {handle.sandbox_id} in FAILED state."
            )


__all__ = ["AgentsSandboxClient"]



from __future__ import annotations

import asyncio
from collections.abc import Iterator
import inspect
from pathlib import Path
from types import SimpleNamespace

import agents_sandbox
import pytest
from agents_sandbox import (
    AgentsSandboxClient,
    CopySpec,
    DeleteSandboxesResult,
    ExecHandle,
    ExecState,
    HealthcheckConfig,
    MountSpec,
    SandboxEvent,
    SandboxEventType,
    SandboxHandle,
    SandboxState,
    ServiceSpec,
)
from agents_sandbox._generated import service_pb2
from agents_sandbox.client import _resolve_default_socket_path
from agents_sandbox._conversions import to_exec_handle, to_exec_snapshot, to_sandbox_handle

from tests.smoke_support import (
    _RecordingSandboxService,
    _event_pb,
    _exercise_public_client,
    _exec_response,
    _legacy_sdk_type_names,
    _new_client,
    _running_server,
    _underscored_name,
)

def test_package_root_exports_only_formal_client() -> None:
    exports = set(getattr(agents_sandbox, "__all__", []))

    assert "AgentsSandboxClient" in exports
    assert "DeleteSandboxesResult" in exports
    assert sorted(name for name in exports if name.endswith("SandboxClient")) == ["AgentsSandboxClient"]
    assert "CreateSandboxRequest" not in exports
    assert "CreateExecRequest" not in exports
    assert "SandboxOwner" not in exports


def test_public_models_match_protocol_contract() -> None:
    service = ServiceSpec(
        name="postgres",
        image="postgres:16",
        environment={"POSTGRES_DB": "agents"},
        healthcheck=HealthcheckConfig(
            test=("CMD-SHELL", "pg_isready -U postgres"),
            interval="5s",
            retries=3,
        ),
        post_start_on_primary=("python", "-c", "print('seeded')"),
    )
    mount = MountSpec(source="/tmp/workspace", target="/workspace", writable=True)
    copy = CopySpec(source="/tmp/source", target="/workspace/source", exclude_patterns=(".git",))

    assert service.name == "postgres"
    assert service.environment["POSTGRES_DB"] == "agents"
    assert service.healthcheck is not None
    assert service.healthcheck.test == ("CMD-SHELL", "pg_isready -U postgres")
    assert mount.target == "/workspace"
    assert copy.exclude_patterns == (".git",)
    assert set(SandboxEvent.__annotations__) == {
        "event_id",
        "sequence",
        "sandbox_id",
        "event_type",
        "occurred_at",
        "replay",
        "snapshot",
        "phase",
        "service_name",
        "error_code",
        "error_message",
        "reason",
        "exec_id",
        "exit_code",
        "sandbox_state",
        "exec_state",
    }
    assert set(SandboxHandle.__annotations__) == {
        "sandbox_id",
        "state",
        "last_event_sequence",
        "required_services",
        "optional_services",
        "labels",
    }
    assert SandboxHandle.__annotations__["last_event_sequence"] == "int"
    assert set(ExecHandle.__annotations__) == {
        "exec_id",
        "sandbox_id",
        "state",
        "command",
        "cwd",
        "env_overrides",
        "exit_code",
        "error",
        "stdout_log_path",
        "stderr_log_path",
        "last_event_sequence",
    }
    assert ExecHandle.__annotations__["last_event_sequence"] == "int"
    assert DeleteSandboxesResult.__annotations__ == {
        "deleted_sandbox_ids": "tuple[str, ...]",
        "deleted_count": "int",
    }



def test_sdk_exports_proto_backed_public_enums() -> None:
    assert SandboxState(service_pb2.SANDBOX_STATE_READY) is SandboxState.READY
    assert ExecState(service_pb2.EXEC_STATE_FINISHED) is ExecState.FINISHED
    assert ExecState.FINISHED.is_terminal is True
    assert ExecState.RUNNING.is_terminal is False
    assert {
        service_pb2.SANDBOX_READY: "sandbox_ready",
        service_pb2.SANDBOX_FAILED: "sandbox_failed",
        service_pb2.EXEC_STARTED: "exec_started",
        service_pb2.EXEC_FINISHED: "exec_finished",
        service_pb2.EXEC_FAILED: "exec_failed",
        service_pb2.EXEC_CANCELLED: "exec_cancelled",
        service_pb2.SANDBOX_SERVICE_READY: "sandbox_service_ready",
        service_pb2.SANDBOX_SERVICE_FAILED: "sandbox_service_failed",
    } == {
        3: "sandbox_ready",
        4: "sandbox_failed",
        9: "exec_started",
        10: "exec_finished",
        11: "exec_failed",
        12: "exec_cancelled",
        13: "sandbox_service_ready",
        14: "sandbox_service_failed",
    }
    assert SandboxEventType(service_pb2.EXEC_FINISHED) is SandboxEventType.EXEC_FINISHED
    assert (
        SandboxEventType(service_pb2.SANDBOX_SERVICE_READY)
        is SandboxEventType.SANDBOX_SERVICE_READY
    )


def test_public_root_exports_remove_legacy_sdk_types() -> None:
    exports = set(getattr(agents_sandbox, "__all__", []))

    for legacy_name in _legacy_sdk_type_names():
        assert legacy_name not in exports


def test_to_exec_handle_maps_exit_code_and_sequence() -> None:
    running = to_exec_handle(
        service_pb2.ExecStatus(
            exec_id="exec-running",
            sandbox_id="sandbox-1",
            state=service_pb2.EXEC_STATE_RUNNING,
            command=["echo", "hello"],
            cwd="/workspace",
            exit_code=0,
            last_event_sequence=3,
        )
    )
    finished = to_exec_handle(
        service_pb2.ExecStatus(
            exec_id="exec-finished",
            sandbox_id="sandbox-1",
            state=service_pb2.EXEC_STATE_FINISHED,
            command=["echo", "hello"],
            cwd="/workspace",
            exit_code=7,
            last_event_sequence=7,
        )
    )

    assert running.exit_code is None  # not terminal, exit_code suppressed
    assert running.last_event_sequence == 3
    assert finished.exit_code == 7
    assert finished.last_event_sequence == 7


def test_to_exec_snapshot_requires_daemon_issued_sequence() -> None:
    valid_exec_status = service_pb2.ExecStatus(
        exec_id="exec-1",
        sandbox_id="sandbox-1",
        state=service_pb2.EXEC_STATE_RUNNING,
        command=["echo", "hello"],
        cwd="/workspace",
    )

    handle, sequence = to_exec_snapshot(_exec_response(valid_exec_status, last_event_sequence=7))

    assert handle.exec_id == "exec-1"
    assert sequence == 7

    zero_sequence_exec_status = service_pb2.ExecStatus(
        exec_id="exec-1",
        sandbox_id="sandbox-1",
        state=service_pb2.EXEC_STATE_RUNNING,
        command=["echo", "hello"],
        cwd="/workspace",
    )

    with pytest.raises(ValueError, match="Sequence must be positive: 0"):
        to_exec_snapshot(SimpleNamespace(exec=zero_sequence_exec_status))

    with pytest.raises(ValueError, match="Sequence must be positive: 0"):
        to_exec_snapshot(_exec_response(zero_sequence_exec_status, last_event_sequence=0))


def test_default_socket_path_resolution_matches_daemon_rules(monkeypatch: pytest.MonkeyPatch) -> None:
    ignored_env_key = "UNRELATED_RUNTIME_SETTING"
    assert _resolve_default_socket_path(
        system="Linux",
        lookup_env=lambda key: {
            ignored_env_key: "/tmp/ignored.sock",
            "XDG_RUNTIME_DIR": "/tmp/runtime",
        }.get(key),
    ) == "/tmp/runtime/agbox/agboxd.sock"
    assert _resolve_default_socket_path(
        system="Darwin",
        lookup_env=lambda _key: None,
        home_dir="/Users/tester",
    ) == "/Users/tester/Library/Application Support/agbox/run/agboxd.sock"
    with pytest.raises(RuntimeError, match="XDG_RUNTIME_DIR"):
        _resolve_default_socket_path(
            system="Linux",
            lookup_env=lambda _key: None,
        )


def test_agents_sandbox_client_signatures_match_public_contract() -> None:
    expected = {
        "ping",
        "create_sandbox",
        "get_sandbox",
        "list_sandboxes",
        "subscribe_sandbox_events",
        "resume_sandbox",
        "stop_sandbox",
        "delete_sandbox",
        "delete_sandboxes",
        "create_exec",
        "run",
        "cancel_exec",
        "get_exec",
        "list_active_execs",
    }
    actual = {
        name
        for name, value in inspect.getmembers(AgentsSandboxClient)
        if callable(value) and not name.startswith("_")
    }

    assert expected - actual == set()
    assert {name for name in actual - expected if name != "close"} == set()

    defaults = {
        "create_sandbox": True,
        "resume_sandbox": True,
        "stop_sandbox": True,
        "delete_sandbox": True,
        "create_exec": False,
        "cancel_exec": True,
    }
    for method_name, expected_wait in defaults.items():
        signature = inspect.signature(getattr(AgentsSandboxClient, method_name))
        assert signature.parameters["wait"].default is expected_wait

    create_signature = inspect.signature(AgentsSandboxClient.create_sandbox)
    assert list(create_signature.parameters) == [
        "self",
        "config",
        "image",
        "sandbox_id",
        "mounts",
        "copies",
        "builtin_resources",
        "required_services",
        "optional_services",
        "labels",
        "wait",
    ]
    assert "request" not in create_signature.parameters
    assert "socket_path" not in inspect.signature(AgentsSandboxClient).parameters
    list_signature = inspect.signature(AgentsSandboxClient.list_sandboxes)
    assert list(list_signature.parameters) == ["self", "include_deleted", "label_selector"]
    delete_many_signature = inspect.signature(AgentsSandboxClient.delete_sandboxes)
    assert list(delete_many_signature.parameters) == ["self", "label_selector", "wait"]
    assert delete_many_signature.parameters["wait"].default is True
    exec_signature = inspect.signature(AgentsSandboxClient.create_exec)
    assert list(exec_signature.parameters) == ["self", "sandbox_id", "command", "exec_id", "cwd", "env_overrides", "wait"]
    assert exec_signature.parameters["cwd"].default == "/workspace"
    assert exec_signature.parameters["exec_id"].default is None
    assert exec_signature.parameters["env_overrides"].default is None
    run_signature = inspect.signature(AgentsSandboxClient.run)
    assert list(run_signature.parameters) == ["self", "sandbox_id", "command", "cwd", "env_overrides"]

    subscribe_signature = inspect.signature(AgentsSandboxClient.subscribe_sandbox_events)
    assert subscribe_signature.parameters["from_sequence"].default == 0
    assert subscribe_signature.parameters["include_current_snapshot"].default is False


def test_public_docs_use_converged_python_sdk_api() -> None:
    repo_root = Path(__file__).resolve().parents[3]
    legacy_socket_token = "LEGACY_SOCKET_ENV"
    public_docs = {
        "README.md": (repo_root / "README.md").read_text(),
        "docs/sdk_async_usage.md": (repo_root / "docs" / "sdk_async_usage.md").read_text(),
        "examples/codex-cli/README.md": (repo_root / "examples" / "codex-cli" / "README.md").read_text(),
    }

    banned_tokens = (
        "CreateSandboxRequest",
        "CreateExecRequest",
        "SandboxOwner",
        *_legacy_sdk_type_names(),
        _underscored_name("SANDBOX", "DEPENDENCY", "READY"),
        'AgentsSandboxClient("/',
        "AgentsSandboxClient('/",
    )

    for relative_path, content in public_docs.items():
        for token in banned_tokens:
            assert token not in content, f"{relative_path} should not mention {token}"
        assert legacy_socket_token not in content, f"{relative_path} should not mention {legacy_socket_token}"

    assert "AgentsSandboxClient()" in public_docs["README.md"]
    assert "AgentsSandboxClient()" in public_docs["docs/sdk_async_usage.md"]
    assert "AgentsSandboxClient()" in public_docs["examples/codex-cli/README.md"]
    assert "ServiceSpec" in public_docs["docs/sdk_async_usage.md"]
    assert "HealthcheckConfig" in public_docs["docs/sdk_async_usage.md"]
    assert "required_services" in public_docs["docs/sdk_async_usage.md"]
    assert "optional_services" in public_docs["docs/sdk_async_usage.md"]
    assert "sandbox_id=" in public_docs["docs/sdk_async_usage.md"]
    assert "exec_id=" in public_docs["docs/sdk_async_usage.md"]
    assert "from_sequence=0" in public_docs["docs/sdk_async_usage.md"]
    assert "last_event_sequence" in public_docs["docs/sdk_async_usage.md"]
    assert "sequence" in public_docs["docs/sdk_async_usage.md"]


def test_public_async_client_round_trips_over_unix_socket(tmp_path: Path) -> None:
    servicer = _RecordingSandboxService()

    with _running_server(tmp_path / "sandbox.sock", servicer):
        result = asyncio.run(_exercise_public_client(tmp_path / "sandbox.sock"))

    assert result["ping"].daemon == "agboxd"
    assert result["sandbox_state"] is SandboxState.READY
    assert result["exec_state"] is ExecState.FINISHED
    assert result["event_types"] == [
        SandboxEventType.SANDBOX_SERVICE_READY,
        SandboxEventType.EXEC_FINISHED,
    ]
    assert result["service_names"] == ["postgres", None]
    assert result["exec_last_event_sequence"] == 3
    assert servicer.subscribe_requests[0].from_sequence == 0
    assert servicer.create_requests[0].create_spec.image == "python:3.12-slim"
    assert servicer.create_requests[0].sandbox_id == "sandbox-1"
    assert dict(servicer.create_requests[0].create_spec.labels) == {"team": "sdk", "purpose": "smoke"}
    assert servicer.create_requests[0].create_spec.required_services == [
        service_pb2.ServiceSpec(
            name="postgres",
            image="postgres:16",
            environment=[service_pb2.KeyValue(key="POSTGRES_DB", value="agents")],
            healthcheck=service_pb2.HealthcheckConfig(
                test=["CMD-SHELL", "pg_isready -U postgres"],
                interval="5s",
                retries=5,
            ),
            post_start_on_primary=["python", "-c", "print('seeded')"],
        )
    ]
    assert servicer.create_requests[0].create_spec.optional_services == [
        service_pb2.ServiceSpec(
            name="redis",
            image="redis:7",
        )
    ]
    assert dict(servicer.list_requests[0].label_selector) == {"team": "sdk"}
    assert dict(servicer.delete_sandboxes_requests[0].label_selector) == {"team": "sdk"}
    assert servicer.create_exec_requests[0].cwd == "/workspace"
    assert result["deleted_count"] == 1


def test_create_sandbox_sandbox_id_serializes_explicit_and_omitted_values(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeRawSandboxClient:
        create_requests: list[service_pb2.CreateSandboxRequest] = []

        def __init__(self, socket_path: str, *, timeout_seconds: float) -> None:
            self.socket_path = socket_path
            self.timeout_seconds = timeout_seconds

        def close(self) -> None:
            return None

        def create_sandbox(self, request: service_pb2.CreateSandboxRequest) -> service_pb2.CreateSandboxResponse:
            self.create_requests.append(request)
            return service_pb2.CreateSandboxResponse(
                sandbox_id=f"sandbox-{len(self.create_requests)}",
                initial_state=service_pb2.SANDBOX_STATE_PENDING,
            )

        def get_sandbox(self, sandbox_id: str) -> service_pb2.GetSandboxResponse:
            return service_pb2.GetSandboxResponse(
                sandbox=service_pb2.SandboxHandle(
                    sandbox_id=sandbox_id,
                    state=service_pb2.SANDBOX_STATE_READY,
                    last_event_sequence=1,
                )
            )

    monkeypatch.setattr("agents_sandbox.client.SandboxGrpcClient", _FakeRawSandboxClient)

    async def run_test() -> None:
        client = _new_client("/tmp/agents-sandbox.sock")
        await client.create_sandbox(image="python:3.12-slim", sandbox_id="sandbox-explicit")
        await client.create_sandbox(image="python:3.12-slim")

    asyncio.run(run_test())

    assert _FakeRawSandboxClient.create_requests[0].sandbox_id == "sandbox-explicit"
    assert _FakeRawSandboxClient.create_requests[1].sandbox_id == ""


def test_create_sandbox_mounts_copies_builtin_resources_and_services_serialize_to_proto(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    class _FakeRawSandboxClient:
        create_requests: list[service_pb2.CreateSandboxRequest] = []

        def __init__(self, socket_path: str, *, timeout_seconds: float) -> None:
            self.socket_path = socket_path
            self.timeout_seconds = timeout_seconds

        def close(self) -> None:
            return None

        def create_sandbox(self, request: service_pb2.CreateSandboxRequest) -> service_pb2.CreateSandboxResponse:
            self.create_requests.append(request)
            return service_pb2.CreateSandboxResponse(
                sandbox_id="sandbox-1",
                initial_state=service_pb2.SANDBOX_STATE_PENDING,
            )

        def get_sandbox(self, sandbox_id: str) -> service_pb2.GetSandboxResponse:
            return service_pb2.GetSandboxResponse(
                sandbox=service_pb2.SandboxHandle(
                    sandbox_id=sandbox_id,
                    state=service_pb2.SANDBOX_STATE_READY,
                    last_event_sequence=1,
                )
            )

    monkeypatch.setattr("agents_sandbox.client.SandboxGrpcClient", _FakeRawSandboxClient)

    async def run_test() -> None:
        client = _new_client("/tmp/agents-sandbox.sock")
        await client.create_sandbox(
            image="python:3.12-slim",
            mounts=(
                MountSpec(source="/host/workspace", target="/workspace", writable=True),
            ),
            copies=(
                CopySpec(
                    source="/host/template",
                    target="/workspace/template",
                    exclude_patterns=(".git", "__pycache__"),
                ),
            ),
            builtin_resources=(".claude", "uv"),
            required_services=(
                ServiceSpec(
                    name="postgres",
                    image="postgres:16",
                    environment={"POSTGRES_DB": "agents"},
                    healthcheck=HealthcheckConfig(
                        test=("CMD-SHELL", "pg_isready -U postgres"),
                        interval="5s",
                        retries=3,
                    ),
                    post_start_on_primary=("python", "-c", "print('seeded')"),
                ),
            ),
            optional_services=(
                ServiceSpec(
                    name="redis",
                    image="redis:7",
                ),
            ),
        )

    asyncio.run(run_test())

    create_spec = _FakeRawSandboxClient.create_requests[0].create_spec
    assert list(create_spec.builtin_resources) == [".claude", "uv"]
    assert create_spec.mounts == [
        service_pb2.MountSpec(source="/host/workspace", target="/workspace", writable=True)
    ]
    assert create_spec.copies == [
        service_pb2.CopySpec(
            source="/host/template",
            target="/workspace/template",
            exclude_patterns=[".git", "__pycache__"],
        )
    ]
    assert create_spec.required_services == [
        service_pb2.ServiceSpec(
            name="postgres",
            image="postgres:16",
            environment=[service_pb2.KeyValue(key="POSTGRES_DB", value="agents")],
            healthcheck=service_pb2.HealthcheckConfig(
                test=["CMD-SHELL", "pg_isready -U postgres"],
                interval="5s",
                retries=3,
            ),
            post_start_on_primary=["python", "-c", "print('seeded')"],
        )
    ]
    assert create_spec.optional_services == [
        service_pb2.ServiceSpec(
            name="redis",
            image="redis:7",
        )
    ]


def test_create_sandbox_with_labels_serializes_to_proto(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeRawSandboxClient:
        create_requests: list[service_pb2.CreateSandboxRequest] = []

        def __init__(self, socket_path: str, *, timeout_seconds: float) -> None:
            self.socket_path = socket_path
            self.timeout_seconds = timeout_seconds

        def close(self) -> None:
            return None

        def create_sandbox(self, request: service_pb2.CreateSandboxRequest) -> service_pb2.CreateSandboxResponse:
            self.create_requests.append(request)
            return service_pb2.CreateSandboxResponse(
                sandbox_id="sandbox-1",
                initial_state=service_pb2.SANDBOX_STATE_PENDING,
            )

        def get_sandbox(self, sandbox_id: str) -> service_pb2.GetSandboxResponse:
            return service_pb2.GetSandboxResponse(
                sandbox=service_pb2.SandboxHandle(
                    sandbox_id=sandbox_id,
                    state=service_pb2.SANDBOX_STATE_READY,
                    last_event_sequence=1,
                    labels={"team": "sdk", "purpose": "test"},
                )
            )

    monkeypatch.setattr("agents_sandbox.client.SandboxGrpcClient", _FakeRawSandboxClient)

    async def run_test() -> None:
        client = _new_client("/tmp/agents-sandbox.sock")
        await client.create_sandbox(
            image="python:3.12-slim",
            labels={"team": "sdk", "purpose": "test"},
        )

    asyncio.run(run_test())

    assert dict(_FakeRawSandboxClient.create_requests[0].create_spec.labels) == {"team": "sdk", "purpose": "test"}


def test_list_sandboxes_with_label_selector(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeRawSandboxClient:
        list_requests: list[service_pb2.ListSandboxesRequest] = []

        def __init__(self, socket_path: str, *, timeout_seconds: float) -> None:
            self.socket_path = socket_path
            self.timeout_seconds = timeout_seconds

        def close(self) -> None:
            return None

        def list_sandboxes(self, request: service_pb2.ListSandboxesRequest) -> service_pb2.ListSandboxesResponse:
            self.list_requests.append(request)
            return service_pb2.ListSandboxesResponse(
                sandboxes=[
                    service_pb2.SandboxHandle(
                        sandbox_id="sandbox-1",
                        state=service_pb2.SANDBOX_STATE_READY,
                        last_event_sequence=1,
                        labels={"team": "sdk"},
                    )
                ]
            )

    monkeypatch.setattr("agents_sandbox.client.SandboxGrpcClient", _FakeRawSandboxClient)

    async def run_test() -> list[SandboxHandle]:
        client = _new_client("/tmp/agents-sandbox.sock")
        return await client.list_sandboxes(include_deleted=True, label_selector={"team": "sdk"})

    sandboxes = asyncio.run(run_test())

    assert _FakeRawSandboxClient.list_requests == [
        service_pb2.ListSandboxesRequest(include_deleted=True, label_selector={"team": "sdk"})
    ]
    assert sandboxes[0].labels == {"team": "sdk"}


def test_delete_sandboxes_waits_for_deleted_state(monkeypatch: pytest.MonkeyPatch) -> None:
    class _FakeRawSandboxClient:
        delete_requests: list[service_pb2.DeleteSandboxesRequest] = []
        subscribe_requests: list[tuple[str, int, bool]] = []

        def __init__(self, socket_path: str, *, timeout_seconds: float) -> None:
            self.socket_path = socket_path
            self.timeout_seconds = timeout_seconds
            self._sandbox_reads: dict[str, int] = {"sandbox-1": 0, "sandbox-2": 0}

        def close(self) -> None:
            return None

        def delete_sandboxes(self, request: service_pb2.DeleteSandboxesRequest) -> service_pb2.DeleteSandboxesResponse:
            self.delete_requests.append(request)
            return service_pb2.DeleteSandboxesResponse(
                deleted_sandbox_ids=["sandbox-1", "sandbox-2"],
                deleted_count=2,
            )

        def get_sandbox(self, sandbox_id: str) -> service_pb2.GetSandboxResponse:
            self._sandbox_reads[sandbox_id] += 1
            state = service_pb2.SANDBOX_STATE_DELETING
            sequence = 1
            if self._sandbox_reads[sandbox_id] >= 2:
                state = service_pb2.SANDBOX_STATE_DELETED
                sequence = 2
            return service_pb2.GetSandboxResponse(
                sandbox=service_pb2.SandboxHandle(
                    sandbox_id=sandbox_id,
                    state=state,
                    last_event_sequence=sequence,
                )
            )

        def subscribe_sandbox_events(
            self,
            sandbox_id: str,
            *,
            from_sequence: int = 0,
            include_current_snapshot: bool = False,
        ) -> Iterator[service_pb2.SandboxEvent]:
            self.subscribe_requests.append((sandbox_id, from_sequence, include_current_snapshot))
            yield _event_pb(
                sandbox_id=sandbox_id,
                sequence=2,
                event_type=service_pb2.SANDBOX_DELETED,
                sandbox_state=service_pb2.SANDBOX_STATE_DELETED,
            )

    monkeypatch.setattr("agents_sandbox.client.SandboxGrpcClient", _FakeRawSandboxClient)

    async def run_test() -> DeleteSandboxesResult:
        client = _new_client("/tmp/agents-sandbox.sock")
        return await client.delete_sandboxes({"team": "sdk"}, wait=True)

    result = asyncio.run(run_test())

    assert isinstance(result, DeleteSandboxesResult)
    assert result.deleted_sandbox_ids == ("sandbox-1", "sandbox-2")
    assert result.deleted_count == 2
    assert _FakeRawSandboxClient.delete_requests == [
        service_pb2.DeleteSandboxesRequest(label_selector={"team": "sdk"})
    ]
    assert _FakeRawSandboxClient.subscribe_requests == [
        ("sandbox-1", 1, False),
        ("sandbox-2", 1, False),
    ]


def test_sandbox_handle_labels_round_trip() -> None:
    handle = to_sandbox_handle(
        service_pb2.SandboxHandle(
            sandbox_id="sandbox-1",
            state=service_pb2.SANDBOX_STATE_READY,
            last_event_sequence=1,
            labels={"team": "sdk"},
        )
    )

    assert handle.labels == {"team": "sdk"}



"""Python client for the agents-sandbox daemon."""

from __future__ import annotations

from collections.abc import Iterator

import grpc
from google.rpc import error_details_pb2
from grpc_status import rpc_status

from . import errors
from ._generated import service_pb2, service_pb2_grpc

_REASON_TO_ERROR = dict(
    SANDBOX_CONFLICT=errors.SandboxConflictError,
    SANDBOX_ID_ALREADY_EXISTS=errors.SandboxConflictError,
    EXEC_ID_ALREADY_EXISTS=errors.SandboxConflictError,
    SANDBOX_NOT_FOUND=errors.SandboxNotFoundError,
    SANDBOX_NOT_READY=errors.SandboxNotReadyError,
    SANDBOX_INVALID_STATE=errors.SandboxInvalidStateError,
    EXEC_NOT_FOUND=errors.ExecNotFoundError,
    EXEC_ALREADY_TERMINAL=errors.ExecAlreadyTerminalError,
    SANDBOX_EVENT_SEQUENCE_EXPIRED=errors.SandboxSequenceExpiredError,
)


class SandboxGrpcClient:
    """Thin synchronous wrapper around the generated gRPC stub."""

    def __init__(self, socket_path: str, *, timeout_seconds: float | None = 5.0) -> None:
        self._timeout_seconds = timeout_seconds
        self._channel = grpc.insecure_channel(f"unix://{socket_path}")
        self._stub = service_pb2_grpc.SandboxServiceStub(self._channel)

    def close(self) -> None:
        self._channel.close()

    def __enter__(self) -> "SandboxGrpcClient":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        del exc_type, exc, tb
        self.close()

    def ping(self) -> service_pb2.PingResponse:
        return self._call(self._stub.Ping, service_pb2.PingRequest())

    def create_sandbox(self, request: service_pb2.CreateSandboxRequest) -> service_pb2.CreateSandboxResponse:
        return self._call(self._stub.CreateSandbox, request)

    def get_sandbox(self, sandbox_id: str) -> service_pb2.GetSandboxResponse:
        return self._call(self._stub.GetSandbox, service_pb2.GetSandboxRequest(sandbox_id=sandbox_id))

    def list_sandboxes(self, request: service_pb2.ListSandboxesRequest | None = None) -> service_pb2.ListSandboxesResponse:
        return self._call(self._stub.ListSandboxes, request or service_pb2.ListSandboxesRequest())

    def resume_sandbox(self, sandbox_id: str) -> service_pb2.AcceptedResponse:
        return self._call(self._stub.ResumeSandbox, service_pb2.ResumeSandboxRequest(sandbox_id=sandbox_id))

    def stop_sandbox(
        self,
        sandbox_id: str,
    ) -> service_pb2.AcceptedResponse:
        return self._call(
            self._stub.StopSandbox,
            service_pb2.StopSandboxRequest(sandbox_id=sandbox_id),
        )

    def delete_sandbox(
        self,
        sandbox_id: str,
    ) -> service_pb2.AcceptedResponse:
        return self._call(
            self._stub.DeleteSandbox,
            service_pb2.DeleteSandboxRequest(sandbox_id=sandbox_id),
        )

    def delete_sandboxes(
        self,
        request: service_pb2.DeleteSandboxesRequest,
    ) -> service_pb2.DeleteSandboxesResponse:
        return self._call(self._stub.DeleteSandboxes, request)

    def subscribe_sandbox_events(
        self,
        sandbox_id: str,
        *,
        from_sequence: int = 0,
        include_current_snapshot: bool = False,
    ) -> Iterator[service_pb2.SandboxEvent]:
        request = service_pb2.SubscribeSandboxEventsRequest(
            sandbox_id=sandbox_id,
            from_sequence=from_sequence,
            include_current_snapshot=include_current_snapshot,
        )
        try:
            stream = self._stub.SubscribeSandboxEvents(request, timeout=self._timeout_seconds)
            for event in stream:
                yield event
        except grpc.RpcError as exc:
            raise _translate_rpc_error(exc) from exc

    def create_exec(self, request: service_pb2.CreateExecRequest) -> service_pb2.CreateExecResponse:
        return self._call(self._stub.CreateExec, request)

    def cancel_exec(
        self,
        exec_id: str,
    ) -> service_pb2.AcceptedResponse:
        return self._call(
            self._stub.CancelExec,
            service_pb2.CancelExecRequest(exec_id=exec_id),
        )

    def get_exec(self, exec_id: str) -> service_pb2.GetExecResponse:
        return self._call(self._stub.GetExec, service_pb2.GetExecRequest(exec_id=exec_id))

    def list_active_execs(self, sandbox_id: str = "") -> service_pb2.ListActiveExecsResponse:
        return self._call(self._stub.ListActiveExecs, service_pb2.ListActiveExecsRequest(sandbox_id=sandbox_id))

    def _call(self, rpc, request):
        try:
            return rpc(request, timeout=self._timeout_seconds)
        except grpc.RpcError as exc:
            raise _translate_rpc_error(exc) from exc


def _translate_rpc_error(exc: grpc.RpcError) -> Exception:
    reason = ""
    parsed_status = rpc_status.from_call(exc)
    if parsed_status is not None:
        for detail in parsed_status.details:
            error_info = error_details_pb2.ErrorInfo()
            if detail.Is(error_info.DESCRIPTOR):
                detail.Unpack(error_info)
                reason = error_info.reason
                break
    error_type = _REASON_TO_ERROR.get(reason, errors.SandboxClientError)
    return error_type(exc.details() or reason or "RPC failed")



"""Public SDK request, response, and handle types."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass, field
from datetime import datetime

from .models import ExecState, SandboxEventType, SandboxState


@dataclass(frozen=True, slots=True)
class PingInfo:
    version: str
    daemon: str


@dataclass(frozen=True, slots=True)
class HealthcheckConfig:
    test: tuple[str, ...]
    interval: str | None = None
    timeout: str | None = None
    retries: int | None = None
    start_period: str | None = None
    start_interval: str | None = None


@dataclass(frozen=True, slots=True)
class ServiceSpec:
    name: str
    image: str
    environment: Mapping[str, str] = field(default_factory=dict)
    healthcheck: HealthcheckConfig | None = None
    post_start_on_primary: tuple[str, ...] = ()


@dataclass(frozen=True, slots=True)
class MountSpec:
    source: str
    target: str
    writable: bool = False


@dataclass(frozen=True, slots=True)
class CopySpec:
    source: str
    target: str
    exclude_patterns: tuple[str, ...] = ()


@dataclass(frozen=True, slots=True)
class SandboxEvent:
    event_id: str
    sequence: int
    sandbox_id: str
    event_type: SandboxEventType
    occurred_at: datetime
    replay: bool = False
    snapshot: bool = False
    phase: str | None = None
    service_name: str | None = None
    error_code: str | None = None
    error_message: str | None = None
    reason: str | None = None
    exec_id: str | None = None
    exit_code: int | None = None
    sandbox_state: SandboxState | None = None
    exec_state: ExecState | None = None


@dataclass(frozen=True, slots=True)
class SandboxHandle:
    sandbox_id: str
    state: SandboxState
    last_event_sequence: int = 0
    required_services: tuple[ServiceSpec, ...] = ()
    optional_services: tuple[ServiceSpec, ...] = ()
    labels: Mapping[str, str] = field(default_factory=dict)

    def __post_init__(self) -> None:
        object.__setattr__(self, "labels", dict(self.labels))


@dataclass(frozen=True, slots=True)
class DeleteSandboxesResult:
    deleted_sandbox_ids: tuple[str, ...]
    deleted_count: int


@dataclass(frozen=True, slots=True)
class ExecHandle:
    exec_id: str
    sandbox_id: str
    state: ExecState
    command: tuple[str, ...]
    cwd: str | None
    env_overrides: Mapping[str, str]
    exit_code: int | None = None
    error: str | None = None
    last_event_sequence: int = 0
    stdout_log_path: str | None = None
    stderr_log_path: str | None = None


__all__ = [
    "CopySpec",
    "DeleteSandboxesResult",
    "ExecHandle",
    "HealthcheckConfig",
    "MountSpec",
    "PingInfo",
    "SandboxEvent",
    "SandboxHandle",
    "ServiceSpec",
]



"""SDK error types."""

from __future__ import annotations


class SandboxClientError(RuntimeError):
    """Base error for SDK failures."""


class SandboxConflictError(SandboxClientError):
    """Raised when a caller-provided ID already exists."""

    def __init__(self, sandbox_id: str | None = None):
        if sandbox_id is None:
            message = "Sandbox already exists."
        else:
            message = sandbox_id
        super().__init__(message)
        self.sandbox_id = sandbox_id


class SandboxNotFoundError(SandboxClientError):
    """Raised when a sandbox does not exist."""

    def __init__(self, sandbox_id: str):
        if " " in sandbox_id:
            message = sandbox_id
            self.sandbox_id = None
        else:
            message = f"Sandbox {sandbox_id} not found."
            self.sandbox_id = sandbox_id
        super().__init__(message)


class SandboxNotReadyError(SandboxClientError):
    """Raised when a sandbox is not ready for exec creation."""

    def __init__(self, sandbox_id: str):
        if " " in sandbox_id:
            message = sandbox_id
            self.sandbox_id = None
        else:
            message = f"Sandbox {sandbox_id} is not ready."
            self.sandbox_id = sandbox_id
        super().__init__(message)


class SandboxInvalidStateError(SandboxClientError):
    """Raised when an operation is invalid for the current sandbox state."""


class ExecNotFoundError(SandboxClientError):
    """Raised when an exec does not exist."""

    def __init__(self, exec_id: str):
        if " " in exec_id:
            message = exec_id
            self.exec_id = None
        else:
            message = f"Exec {exec_id} not found."
            self.exec_id = exec_id
        super().__init__(message)


class ExecAlreadyTerminalError(SandboxClientError):
    """Raised when an exec is already terminal."""

    def __init__(self, exec_id: str):
        if " " in exec_id:
            message = exec_id
            self.exec_id = None
        else:
            message = f"Exec {exec_id} is already terminal."
            self.exec_id = exec_id
        super().__init__(message)


class ExecNotRunningError(SandboxInvalidStateError):
    """Raised when an exec is no longer running."""

    def __init__(self, exec_id: str):
        if " " in exec_id:
            message = exec_id
            self.exec_id = None
        else:
            message = f"Exec {exec_id} is not running."
            self.exec_id = exec_id
        super().__init__(message)


class SandboxSequenceExpiredError(SandboxClientError):
    """Raised when an event sequence anchor has expired."""

    def __init__(self, sandbox_id: str, from_sequence: int | None = None, oldest_sequence: int | None = None):
        if from_sequence is None or oldest_sequence is None:
            message = sandbox_id
            self.sandbox_id = None
            self.from_sequence = None
            self.oldest_sequence = None
        else:
            message = (
                f"Sandbox {sandbox_id} event sequence {from_sequence} expired; "
                f"oldest retained sequence is {oldest_sequence}."
            )
            self.sandbox_id = sandbox_id
            self.from_sequence = from_sequence
            self.oldest_sequence = oldest_sequence
        super().__init__(message)
